(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *
 * 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************.***************************************************************)
(* $Id: Normalized.thy 9177 2011-05-23 13:19:59Z lukasbru $ *)

theory 
  "Normalized"
imports
  "../../PolicyDefinitions"
begin
text{*
In this theory we explore the effect of policy normalisation on a small policy.*}
definition Policy :: "(IntegerPort net,port) Combinators" where
"Policy = 
DenyAll \<oplus>  AllowPortFromTo two one 683 \<oplus> DenyAllFromTo one three \<oplus> 
DenyAllFromTo four two \<oplus> AllowPortFromTo four one 30"


lemma "dupl (policy2list Policy) = X"
apply (insert nets_different)
apply (simp add: Policy_def,thin_tac "?X")
oops


lemmas policies =  Policy_def 

lemmas UnfoldNetworkAndPolicy = UnfoldPolicy PLemmas policies 
                                 IntegerPort.src_port_def 


lemmas normalizeUnfold = makeFUT_def normalize_def policies Nets_List_def 
                          bothNets_def aux aux2 bothNets_def

lemmas casesSimp = fixDefs packet_Nets_def PLemmas UnfoldNetworkAndPolicy


lemma noMT2: "\<forall> x \<in> set (policy2list Policy). dom (C x) \<noteq> {}"
apply (simp add: UnfoldNetworkAndPolicy)
done

lemma count_the_rules: 
   "(int (length(policy2list (list2policy(normalize Policy)))) = post) \<and> 
    (int(length (policy2list Policy)) = pre) \<and>
    (int (length((normalize Policy))) = Partitions)"
apply (insert nets_different noMT2)
apply (simp add:  normalizeUnfold)
apply (thin_tac "?X")+
oops


lemma normalize: "normalize Policy = X"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold, thin_tac "?X",thin_tac "?S")
oops

text{* The normalisation splits the policy into four parts, which are processed in 
        the following. *}

test_spec "fixElements x \<longrightarrow> makeFUT FUT Policy x 0"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold , thin_tac "?X",thin_tac "?S")
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (simp add: casesSimp)
apply (gen_test_cases "FUT")
store_test_thm "part0"

gen_test_data "part0"

thm part0.test_data


test_spec "fixElements x \<longrightarrow> makeFUT FUT Policy x 1"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold , thin_tac "?X",thin_tac "?S")
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (simp add: casesSimp)
apply (gen_test_cases "FUT")
store_test_thm "part1"

gen_test_data "part1"

test_spec "fixElements x \<longrightarrow> makeFUT FUT Policy x 2"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold , thin_tac "?X",thin_tac "?S")
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (simp add: casesSimp)
apply (gen_test_cases "FUT")
store_test_thm "part2"

gen_test_data "part2"

test_spec "fixElements x \<longrightarrow> makeFUT FUT Policy x 3"
apply (insert nets_different noMT2)
apply (simp add: normalizeUnfold , thin_tac "?X",thin_tac "?S")
apply (rule_tac x=x in spec, rule allI, simp only: split_tupled_all, rule impI)
apply (simp add: casesSimp)
apply (gen_test_cases "FUT")
store_test_thm "part3"

gen_test_data "part3"

text{* The complete set of test data for this policy *}


thm part0.test_data part1.test_data part2.test_data part3.test_data
text{*
FUT (fixID, (2, -3), (1, 683), data) =
Some (accept (fixID, (2, -3), (1, 683), data))

FUT (fixID, (1, -9), (2, 683), data) =
Some (deny (fixID, (1, -9), (2, 683), data))

FUT (fixID, (2, 1), (1, -4), data) =
Some (deny (fixID, (2, 1), (1, -4), data))

FUT (fixID, (1, 3), (2, 8), data) =
Some (deny (fixID, (1, 3), (2, 8), data))

FUT (fixID, (1, 0), (3, -1), data) =
Some (deny (fixID, (1, 0), (3, -1), data))

FUT (fixID, (3, -6), (1, 2), data) =
Some (deny (fixID, (3, -6), (1, 2), data))

FUT (fixID, (4, 5), (2, -2), data) =
Some (deny (fixID, (4, 5), (2, -2), data))

FUT (fixID, (2, -6), (4, 0), data) =
Some (deny (fixID, (2, -6), (4, 0), data))

FUT (fixID, (4, -2), (1, 30), data) =
Some (accept (fixID, (4, -2), (1, 30), data))

FUT (fixID, (1, 8), (4, 30), data) =
Some (deny (fixID, (1, 8), (4, 30), data))

FUT (fixID, (4, -3), (1, -9), data) =
Some (deny (fixID, (4, -3), (1, -9), data))

FUT (fixID, (1, 10), (4, -4), data) =
Some (deny (fixID, (1, 10), (4, -4), data))
*}


text{* And next the standard approach without using normalisation *}

test_spec "fixElements x \<longrightarrow>  FUT x = C Policy x"
apply (prepare_fw_spec)
apply (simp add: fixDefs packet_Nets_def)
apply (simp add: PLemmas  UnfoldNetworkAndPolicy)
apply (gen_test_cases "FUT")
apply simp_all
store_test_thm "full"
gen_test_data "full"

thm full.test_data

text{*
FUT (fixID, (4, 7), (2, -3), data) =
Some (deny (fixID, (4, 7), (2, -3), data))

FUT (fixID, (1, -4), (3, -8), data) =
Some (deny (fixID, (1, -4), (3, -8), data))

FUT (fixID, (2, 6), (-6, -9), data) =
Some (deny (fixID, (2, 6), (-6, -9), data))

FUT (fixID, (5, 1), (-1, 3), data) =
Some (deny (fixID, (5, 1), (-1, 3), data))

FUT (fixID, (-5, -3), (7, 6), data) =
Some (deny (fixID, (-5, -3), (7, 6), data))

FUT (fixID, (-5, -3), (-3, 1), data) =
Some (deny (fixID, (-5, -3), (-3, 1), data))

FUT (fixID, (-3, -4), (2, 3), data) =
Some (deny (fixID, (-3, -4), (2, 3), data))

FUT (fixID, (8, -5), (9, 5), data) =
Some (deny (fixID, (8, -5), (9, 5), data))

FUT (fixID, (2, -2), (1, 683), data) =
Some (accept (fixID, (2, -2), (1, 683), data))

FUT (fixID, (-3, 1), (-4, -6), data) =
Some (deny (fixID, (-3, 1), (-4, -6), data))

FUT (fixID, (-3, 3), (-2, 7), data) =
Some (deny (fixID, (-3, 3), (-2, 7), data))

FUT (fixID, (3, 6), (7, 6), data) =
Some (deny (fixID, (3, 6), (7, 6), data))

FUT (fixID, (-10, 9), (-2, -6), data) =
Some (deny (fixID, (-10, 9), (-2, -6), data))

FUT (fixID, (1, 3), (3, 6), data) =
Some (deny (fixID, (1, 3), (3, 6), data))

FUT (fixID, (-1, -8), (-4, 1), data) =
Some (deny (fixID, (-1, -8), (-4, 1), data))

FUT (fixID, (-4, 3), (-8, -5), data) =
Some (deny (fixID, (-4, 3), (-8, -5), data))

FUT (fixID, (8, 6), (9, -8), data) =
Some (deny (fixID, (8, 6), (9, -8), data))

FUT (fixID, (-3, 4), (-9, -2), data) =
Some (deny (fixID, (-3, 4), (-9, -2), data))

FUT (fixID, (3, 5), (5, 1), data) =
Some (deny (fixID, (3, 5), (5, 1), data))

FUT (fixID, (-9, -5), (-5, 5), data) =
Some (deny (fixID, (-9, -5), (-5, 5), data))

FUT (fixID, (4, 8), (1, 30), data) =
Some (accept (fixID, (4, 8), (1, 30), data))

FUT (fixID, (5, 1), (-7, -3), data) =
Some (deny (fixID, (5, 1), (-7, -3), data))

FUT (fixID, (5, 0), (-2, -6), data) =
Some (deny (fixID, (5, 0), (-2, -6), data))

FUT (fixID, (-2, 5), (-8, 7), data) =
Some (deny (fixID, (-2, 5), (-8, 7), data))

FUT (fixID, (4, -9), (-5, 5), data) =
Some (deny (fixID, (4, -9), (-5, 5), data))

FUT (fixID, (8, 7), (5, 8), data) =
Some (deny (fixID, (8, 7), (5, 8), data))

FUT (fixID, (1, -9), (-10, -1), data) =
Some (deny (fixID, (1, -9), (-10, -1), data))

FUT (fixID, (-3, -7), (8, -3), data) =
Some (deny (fixID, (-3, -7), (8, -3), data))

FUT (fixID, (10, 4), (-1, 5), data) =
Some (deny (fixID, (10, 4), (-1, 5), data))

FUT (fixID, (5, 3), (8, -6), data) =
Some (deny (fixID, (5, 3), (8, -6), data))

FUT (fixID, (2, -6), (-5, 5), data) =
Some (deny (fixID, (2, -6), (-5, 5), data))

FUT (fixID, (-4, -7), (4, 2), data) =
Some (deny (fixID, (-4, -7), (4, 2), data))

FUT (fixID, (-2, -7), (-2, 3), data) =
Some (deny (fixID, (-2, -7), (-2, 3), data))

FUT (fixID, (-4, -9), (10, -2), data) =
Some (deny (fixID, (-4, -9), (10, -2), data))

FUT (fixID, (9, 5), (9, 6), data) =
Some (deny (fixID, (9, 5), (9, 6), data))

FUT (fixID, (8, 7), (0, -9), data) =
Some (deny (fixID, (8, 7), (0, -9), data))

FUT (fixID, (-2, 9), (-9, 7), data) =
Some (deny (fixID, (-2, 9), (-9, 7), data))

FUT (fixID, (-1, -10), (6, -10), data) =
Some (deny (fixID, (-1, -10), (6, -10), data))

FUT (fixID, (-5, -10), (1, -8), data) =
Some (deny (fixID, (-5, -10), (1, -8), data))
  *}

export_test_data "full.data" "full"

end
