theory NormalisationTestSpecification
imports Normalisation
begin



fun disjDom where
 "disjDom (x#xs) = ((\<forall>y\<in>(set xs). dom x \<inter> dom y = {}) \<and> disjDom xs)"
|"disjDom [] = True" 


fun FUTList :: "('a \<mapsto> 'b) \<Rightarrow> 'a \<Rightarrow> ('a \<mapsto> 'b) list \<Rightarrow> bool"
 where
 "FUTList FUT x (p#ps) = ((x \<in> dom p \<longrightarrow> (FUT x = p x)) \<and> (FUTList FUT x ps))"
|"FUTList FUT x [] = True" 


lemma distrFUTL1[rule_format]: 
 "x \<in> dom P \<longrightarrow> (list2policyP PL) x = P x  \<longrightarrow> 
(FUTList FUT x PL \<longrightarrow> (FUT x = P x))"
apply (induct_tac PL) 
apply simp_all
apply (simp add: dom_def)
apply auto
done



lemma FUTList_None[rule_format]: 
 "x \<notin> dom (list2policyP list) \<longrightarrow> FUTList FUT x list"
apply (induct_tac list)
apply simp_all
done



lemma FUTList_DomMT[rule_format]:
  "(\<forall>y\<in>set list. dom a \<inter> dom y = {}) \<longrightarrow> x \<in> (dom a) \<longrightarrow>
        x \<notin> dom (list2policyP list)"
apply (induct_tac list)
apply (simp_all)
apply auto
done


lemma distrFUTL2[rule_format]: 
  "x \<in> dom P \<longrightarrow> (list2policyP PL) x = P x  \<longrightarrow> disjDom PL \<longrightarrow>
(FUT x = P x) \<longrightarrow> FUTList FUT x PL "
apply (induct_tac PL) 
apply simp_all
apply (rule impI)+
apply simp
apply (case_tac "x \<in> dom a")
apply simp_all
defer 1
apply (drule mp)
apply simp_all
apply (metis map_add_dom_app_simps(3))
apply (rule conjI)
apply (drule sym)
apply simp
apply (rule domStart)
apply simp_all
apply (case_tac "list2policyP list x = P x")
apply simp
apply simp
apply (drule sym)
apply simp
apply (rule FUTList_None)
apply (rule_tac a = a in FUTList_DomMT)
apply simp_all
done




lemma distrFUTL[rule_format]: 
  "\<lbrakk>x \<in> dom P; (list2policyP PL) x = P x; disjDom PL\<rbrakk> \<Longrightarrow>  
(FUT x = P x) = FUTList FUT x PL "
apply (rule iffI)
apply (rule distrFUTL2)
apply simp_all
apply (rule_tac PL = PL in distrFUTL1)
apply auto
done



text{* It makes sense to cater for the common special case where the
normalisation returns a list where the last element is a
default-catch-all rule. It seems easier to cater for this globally,
rather than to require the normalisation procedures to do this.  *}


fun gatherDomain_aux where
 "gatherDomain_aux (x#xs) = (dom x \<union> (gatherDomain_aux xs))"
|"gatherDomain_aux [] = {}"

definition gatherDomain where "gatherDomain p = (gatherDomain_aux (butlast p))"


definition FUTListGD where "FUTListGD FUT x p = 
  (((x \<notin> (gatherDomain p) \<and> x \<in> dom (last p)) \<longrightarrow> FUT x = (last p) x) \<and> 
                          (FUTList FUT x (butlast p)))"


definition disjDomGD where "disjDomGD p = disjDom (butlast p)"


lemma disjDom_aux[rule_format]: "disjDom x \<longrightarrow> disjDom (butlast x)"
apply (induct_tac x)
apply simp_all
apply (rule impI)
by (metis in_set_butlastD)


lemma aux1a:  
"(x \<notin> gatherDomain P \<longrightarrow> FUT x = last P x)  \<longrightarrow> 
 FUTList FUT x (butlast P) = FUTListGD FUT x P"
apply (simp add: FUTListGD_def)
done



lemma distrFUTLG1[rule_format]: 
  "x \<in> dom P \<longrightarrow> (list2policyP PL) x = P x \<longrightarrow>  
(FUTListGD FUT x PL \<longrightarrow> (FUT x = P x))"
apply (simp add: FUTListGD_def disjDomGD_def gatherDomain_def)
apply (induct_tac PL)
apply simp
apply (simp_all add: dom_def)
apply (metis option.simps(2))
apply (rule impI)+
apply simp
by (metis domIff gatherDomain_def map_add_dom_app_simps(1) map_add_dom_app_simps(3))





lemma FUTListG_None[rule_format]: 
 "list \<noteq> [] \<longrightarrow> x \<notin> dom (list2policyP list) \<longrightarrow> FUTListGD FUT x list"
apply (simp add: FUTListGD_def )
apply (induct_tac list)
apply simp_all
apply (rule impI)+
apply simp
apply (subgoal_tac "(x \<notin> gatherDomain list)")
apply simp
apply (simp add: gatherDomain_def)
done



lemma distrFUTLG2[rule_format]: 
  "PL \<noteq> [] \<longrightarrow> x \<in> dom P \<longrightarrow> (list2policyP (PL)) x = P x \<longrightarrow> disjDomGD PL \<longrightarrow> 
   (FUT x = P x) \<longrightarrow> FUTListGD FUT x (PL)"
apply (simp add: FUTListGD_def disjDomGD_def gatherDomain_def)
apply (induct_tac PL)
apply simp_all
apply (rule impI)+
apply simp
apply (rule conjI)
apply (rule impI)+
apply simp
apply (metis map_add_dom_app_simps(3))
apply (rule conjI|rule impI)+
apply (metis map_add_dom_app_simps(1))
apply (metis FUTList_DomMT FUTList_None map_add_dom_app_simps(3))
done


lemma distrFUTLG[rule_format]: 
  "\<lbrakk>x \<in> dom P; (list2policyP PL) x = P x; disjDomGD PL; PL \<noteq> []\<rbrakk> \<Longrightarrow>  
(FUT x = P x) = FUTListGD FUT x PL "
apply (rule iffI)
apply (rule distrFUTLG2)
apply simp_all
apply (rule_tac PL = PL in distrFUTLG1)
apply auto
done


end


